#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Python3 / SageMath script to check for weak RSA moduli in a set of
# RSA certificates. Tested with with sage 9.0/9.1 and Python 3.6.9/3.8.5.
#
# The weak moduli in the given list '10000_moduli' are found within 2 sec on a modern PC.
# For 100,000 moduli (each with a size of 512 bit): within 30 sec on a modern 8 core PC.
#
# Version: 1.0 (vm, vs, be, mf) / 2014
# Version: 2.0 (bb, (be)) / 2020.08.21
#
# (C) CrypTool Project [Apache licence and GPL3 and APL]

"""RSA Sanity Check -- Testing RSA moduli for shared prime factors.

Python3 / SageMath script to check a set of RSA certificates for weak moduli, i.e. moduli with shared primes. Use the supplied shell script `moduli_extract.sh` to generate a 'MODULI' file from a set of PEM certificates.

Examples:
    sage -python3 {prog:} -h
    sage -python3 {prog:} -i MODULI
    sage -python3 {prog:} -i MODULI -f CERTFILENAMES -r REPORT

Run 'sage -python3 {prog:} -h' for a complete list of possible options.
Run 'sage -python3 {prog:} -i 10000_moduli' to just check the given moduli (without the need to first extract them from PEM files via the shell script).
"""

import datetime
import os
import sys
from argparse import (ArgumentParser, ArgumentTypeError,
                      RawDescriptionHelpFormatter)
from contextlib import contextmanager
from itertools import islice, zip_longest
from math import ceil
from multiprocessing import Pool

from sage.all import gcd


# ==== UTILS ====
def eprint(*args, **kwargs):
    """Print a message to `stderr`."""
    print(*args, file=sys.stderr, **kwargs)


def print_report(report_file, *args, **kwargs):
    """Print and write a message to `report_file`. Currently only used in main()."""
    kwargs.setdefault("end", "\n")

    print(*args, **kwargs)

    if report_file:
        report_file.write(" ".join(map(str, args)) + kwargs["end"])


def chunked(iterable, n):
    """Return iterator over `n`-item chunks of `iterable`."""
    it = iter(iterable)
    while True:
        chunk = tuple(islice(it, n))
        if not chunk:
            return
        yield chunk


@contextmanager
def nullctx():
    """Return dummy null-context (for python <= 3.6)."""
    yield None


# ==== MODULI CHECKER ====
def load_moduli(moduli_fname, base=16, filenames_fname=None):
    """Load moduli from supplied file."""
    # Load moduli from file (this file may have comment lines staring with "#")
    with open(moduli_fname, 'r') as moduli_f:
        lines = filter(lambda x: x[1] != "",
                       map(lambda x: (x[0], x[1].split("#")[0].strip(" \n")),
                           filter(lambda x: not x[1].startswith("#"),
                                  enumerate(moduli_f.readlines()))))
        moduli = [(int(m, base), line + 1) for line, m in lines]

    if len(moduli) == 0:
        raise RuntimeError(f"No moduli supplied in '{moduli_fname}'!")

    # Load corresponding filenames from file
    filenames = None

    if filenames_fname is not None:
        with open(filenames_fname, 'r') as filenames_f:
            lines = filter(lambda x: x != "",
                           map(lambda x: x.split("#")[0].strip(" \n"),
                               filter(lambda x: not x.startswith("#"),
                                      filenames_f.readlines())))
            filenames = list(lines)

        if len(moduli) != len(filenames):
            raise RuntimeError(f"The number of lines in '{moduli_fname}' and "
                               f"'{filenames_fname}' do not match!")

        moduli = [(mod[0], filenames[idx]) for idx, mod in enumerate(moduli)]

    # Remove duplicate moduli
    seen = set()
    dup_indices = set()
    duplicates = []

    for i, (mod, idx) in enumerate(moduli):
        if mod not in seen:
            seen.add(mod)
        else:
            for org_mod, org_idx in moduli:
                if org_mod == mod:
                    dup_indices.add(i)
                    duplicates.append(((mod, idx), org_idx))
                    break

    for dup in dup_indices:
        del moduli[dup]

    return moduli, duplicates


def task_calc_gcds(args):
    """MP Task: Find shared factor in two consecutive entries."""
    moduli = args
    result = []
    gcds = set()

    # Check two consecutive entries and store their product to be checked in
    # the next iteration
    i = len(moduli)
    for k in range(0, i, 2):
        if k+1 >= i:
            result.append(moduli[k])
        else:
            temp_gcd = gcd(moduli[k], moduli[k+1])

            if temp_gcd != 1:
                gcds.add(temp_gcd)

            result.append(moduli[k] * moduli[k+1])

    return result, gcds


def task_calc_gcds2(args):
    """MP Task: Find shared factor between first and every other entry."""
    mod, candidates = args
    gcds = set()

    # Check mod with every other candidate
    for candidate in candidates:
        temp_gcd = gcd(mod, candidate)
        if temp_gcd != 1:
            gcds.add(temp_gcd)

    return gcds


def task_check_moduli(args):
    """MP Task: Find shared factor in each entry."""
    gcds, moduli = args
    result = set()

    # Find modul that shares prime (index, modul)
    for mod, idx in moduli:
        for temp_gcd in gcds:
            if gcd(mod, temp_gcd) != 1:
                result.add((mod, idx))

    return result


def check_moduli(moduli):
    """Check GCD of all moduli."""
    # Schedule moduli evenly for task runners
    def schedule(data, n):
        for i, mod in enumerate(data[:-1]):
            iterator = chunked(data[i+1:], n)
            while True:
                try:
                    yield (mod, next(iterator))
                except StopIteration:
                    break

    tasks = os.cpu_count()
    print(f" :: Number of cpus used for check_moduli(): {tasks}")
    with Pool(processes=tasks) as p:
        # Copy moduli (without index) into working list
        tree = list(map(lambda x: x[0], moduli))
        gcds = set()

        # Phase one: Find the gcds different than 1
        while len(tree) > tasks * 2:
            results = []
            chunks = max(2, ceil(len(tree) / tasks))
            for res in p.imap(task_calc_gcds, chunked(tree, chunks)):
                results.extend(res[0])
                gcds.update(res[1])
            tree = results

        for res in p.imap(task_calc_gcds2, schedule(tree, tasks)):
            gcds.update(res)

        # Phase two: Trace moduli which share factors
        mod_share = set()

        if gcds:
            chunks = ceil(len(moduli) / tasks)
            iterator = zip_longest([], chunked(moduli, chunks),
                                   fillvalue=gcds)
            for res in p.imap(task_check_moduli, iterator):
                mod_share.update(res)

    return mod_share


def factor_moduli(moduli):
    """Factor weak moduli into shared prime factors."""
    # Copy weak moduli into sorted working list
    moduli = sorted(moduli, key=lambda m: m[1])

    # Phase three: Factor moduli with shared primes
    weak_mod = {}

    for i, (mod1, idx1) in enumerate(moduli):
        # Test mod1 with every other modul mod2 coming before
        for mod2, idx2 in moduli[:i]:
            temp_gcd = int(gcd(mod1, mod2))
            if temp_gcd != 1:
                if mod2 not in weak_mod:
                    weak_mod[mod2] = (idx2, temp_gcd, mod2 // temp_gcd)
                if mod1 not in weak_mod:
                    weak_mod[mod1] = (idx1, temp_gcd, mod1 // temp_gcd)

    return weak_mod


def main(moduli, filenames=None, report=None, base=16):
    """Check supplied moduli for shared primes and generate report."""
    print(f" :: Loading moduli from '{moduli[0]}'...")
    moduli, duplicates = load_moduli(moduli[1], base,
                                     filenames[1] if filenames else None)

    # Start timer
    time_start = datetime.datetime.now().replace(microsecond=0)

    print(f" :: Testing {len(moduli)} RSA moduli for shared prime factors...")
    weak_mod = check_moduli(moduli)

    if weak_mod:
        print(" :: Factor weak moduli into prime factors...")
        weak_mod = factor_moduli(weak_mod)

    # Stop timer
    time_finish = datetime.datetime.now().replace(microsecond=0)

    # Print result
    with (open(report, 'w') if report is not None else nullctx()) as r:
        print_report(r, f"\nTested {len(moduli)} RSA moduli\n")

        if duplicates:
            for (_, dup), org in duplicates:
                duplicate = f"line {dup}" if filenames is None else dup
                original = f"line {org}" if filenames is None else org
                print_report(r, f" -- Duplicate modulus in {duplicate} "
                                f"(duplicate of {original})")

            print_report(r, "")

        if weak_mod:
            indices, factors1, factors2 = zip(*weak_mod.values())
            if filenames is None:
                index = "Line"
                index_width = max(4, max(map(lambda s: len(str(s)), indices)))
            else:
                index = "Filename"
                index_width = max(8, max(map(len, indices)))

            if base == 16:
                modulus_width = max(8, max(map(lambda s: len(hex(s)), weak_mod)))
            else:
                modulus_width = max(8, max(map(lambda s: len(str(s)), weak_mod)))

            factor1_width = max(8, max(map(lambda s: len(str(s)), factors1)))
            factor2_width = max(8, max(map(lambda s: len(str(s)), factors2)))

            header = "    ".join((
                "  {0:<{width}}".format(index, width=index_width),
                "{0:<{width}}".format("Modulus", width=modulus_width),
                "{0:<{width}}".format("Factor 1", width=factor1_width),
                "{0:<{width}}  ".format("Factor 2", width=factor2_width),
            ))
            print_report(r, f"{header}\n{'-' * len(header)}")

            for mod in weak_mod:
                idx, factor1, factor2 = weak_mod[mod]
                print_report(r, "    ".join((
                    f"  {idx:<{index_width}}",
                    f"{mod:>#{modulus_width}x}" if base == 16 else f"{mod:>{modulus_width}}",
                    f"{factor1:>{factor1_width}}",
                    f"{factor2:>{factor2_width}}",
                )))

            print_report(r, f"\n ==> {len(weak_mod)} weak moduli found!")

        else:
            print_report(r, " ==> No weak moduli found.")

        print_report(r, f"\nStarted:  {time_start}\n"
                        f"Finished: {time_finish}\n"
                        f"Processing time: {time_finish - time_start}\n")

    if report is not None:
        print(f"Report written to '{report}'")


# ################# main() #################
if __name__ == '__main__':

    # init argument parser
    doc = [p + "\n" for p in __doc__.split("\n\n")]
    description = "\n".join(doc[0:2]).format(prog=sys.argv[0])
    epilog = doc[2].format(prog=sys.argv[0])

    def _file(fname):
        """Check if file exists."""
        path = os.path.abspath(fname)
        if not os.path.exists(path):
            raise ArgumentTypeError(f"File '{fname}' not found")

        path = os.path.realpath(path)
        if not os.path.isfile(path):
            raise ArgumentTypeError("'{fname}' is not a file")

        return (fname, path)

    # add cli arguments
    parser = ArgumentParser(description=description, epilog=epilog,
                            formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument("-d", "--decimal", dest="base", action="store_const",
                        const=10, default=16,
                        help="moduli are in decimal format. Default are hex encoded moduli")
    parser.add_argument("-i", dest="moduli", type=_file, required=True,
                        help="read file with moduli to check (one per line, default: hex encoded)")
    parser.add_argument("-f", dest="filenames", type=_file, metavar="CERTFILENAMES",
                        help="read file with names of certificate files corresponding to the moduli")
    parser.add_argument("-r", dest="report",
                        help="output report text file")

    # gracefully handle no_args
    if len(sys.argv) == 1:
        eprint(__doc__.format(prog=sys.argv[0]), end="")
        sys.exit(1)

    # run with arguments
    args = parser.parse_args()
    main(**vars(args))
